/*
	wavconv.c
	(c) 2008 A Grosz <grosza@hotmail.com>

	This is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	It is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <io.h>
#include <stdlib.h>
#include <time.h>
#include "pcmwav.h"

#define COPYRIGHT_NOTICE	"wavconv v1.0 (c) 2008 A Grosz <grosza@hotmail.com>.\n" \
							"All rights reserved."

unsigned char			*buf;
size_t			iobufsize = 10000000;
pcmwavfile		pwf;
//HANDLE			outf;
FILE			*outf;
unsigned char	threshold = 0;
int				quiet = 0, nooverwrite = 0;
char			outfname[1024];
int				dowhat = 0, prompt = 0;
int				dontabort = 0;

unsigned long passthrough(void);
int process_filespec(char *fspec);
int process_file(char *fname);
void usage(void);

int main(int argc, char *argv[]) {

	int				i;

	/* Parse command line */
	for (i = 1; i < argc; i++) {
		if ((argv[i][0] == '-') && (argv[i][1] != 0x00)) {
			switch (argv[i][1]) {
				case 'h':
					usage();
					return 0;
				case 'q':
					quiet = 1;
					break;
				case 'd':
					dontabort = 1;
					break;
				case 'p':
					prompt = 1;
					break;
				case 't':
					threshold = atoi(argv[++i]);
					break;

				case 'o':
					nooverwrite = 1;
					strcpy(outfname, argv[++i]);
					break;

				default:
					fprintf(stderr, "Error: Can't understand flag -%c. Aborting.\n", argv[i][1]);
					return 2;
				}
		} else {
			break;
		}
	}

	if (i >= argc) {
		usage();
		return 2;
	}

	if (!quiet)
		fprintf(stderr, "\n%s\n\n", COPYRIGHT_NOTICE);

	return process_filespec(argv[i]);

	return 0;
}

int process_filespec(char *fspec) {
	intptr_t	hFile;
	char	myfullpath[_MAX_PATH];
	char	drive[_MAX_DRIVE];
	char	dir[_MAX_DIR];
	struct	_finddata_t my_file;
	int		err;

	_fullpath(myfullpath, fspec, _MAX_PATH);
	_splitpath(myfullpath, drive, dir, NULL, NULL);

	if ((hFile = _findfirst(fspec, &my_file)) == -1L)
		fprintf(stderr, "Could not find file %s.\n", fspec);
	else {

		do {

			if (my_file.attrib & _A_SUBDIR)
				continue;

			sprintf(myfullpath, "%s%s%s", drive, dir, my_file.name);

			err = process_file(myfullpath);
			if (err && (err != 3)) {

				if ((err != 5) || !dontabort)
					return err;
			}

		} while (_findnext(hFile, &my_file) == 0);

		_findclose(hFile);
	}

	return err;
}

unsigned long passthrough(void)
{
	unsigned long	ndone = 0, readn, i, toWrite, bitcount;
	int				lastn = -1;
	static unsigned int bit = 0;

	readn = (unsigned long) iobufsize;
	if (pwf.ndatabytes < iobufsize)
		readn = pwf.ndatabytes;

	if (!pcmwav_read(&pwf, buf, readn)) {
		if (!quiet)
			fprintf(stderr, "%s\n", pcmwav_error);
		return 0;
	}

	i = 0;
	toWrite = 0;
	bitcount = 1;
	while (i < readn) {
		switch(pwf.bitspersample) {
			case 1:
				{
					unsigned char in;
					unsigned char out;
					unsigned char bit;

					in = *((unsigned char*)buf + i);
					bitcount = 1;
					while (bitcount <= 8) {
						bit = (in >> (8 - bitcount)) & 1;
						out = bit ? 0xD8 : 0x80;
						fputc(out, outf);
						toWrite += 1;
						bitcount += 1;
					}
				}
				break;
			case 8:
				{
					unsigned char byte = *(buf + i);
					unsigned char mythreshold = (128 * threshold) / 100;

					if (byte > 0x80 + mythreshold) {
						bit = 1;
					} else if (byte <= 0x80 - mythreshold) {
						bit = 0;
					}
					bitcount = (bitcount << 1) | bit;
				}
				if (bitcount & 0x100) {
					*((unsigned char*)buf + toWrite) = (unsigned char) (bitcount & 0xFF);
					toWrite += 1;
					bitcount = 1;
				}
				break;
			case 16:
				{
					short byte = *((short*)(buf + i));
					short mythreshold = (16384 * threshold) / 100;

					if (byte > mythreshold) {
						bit = 1;
					} else if (byte <= -mythreshold) {
						bit = 0;
					}
					bitcount = (bitcount << 1) | bit;
				}
				if (bitcount & 0x100) {
					*((unsigned char*)buf + toWrite) = (unsigned char)(bitcount & 0xFF);
					toWrite += 1;
					bitcount = 1;
				}
				break;
		}
		i++;
	}
	if (pwf.bitspersample != 1) {
		while (!(bitcount & 0x100)) {
			bitcount = (bitcount << 1) | 0;
		}
		*((unsigned char*)buf + toWrite++) = (unsigned char)(bitcount & 0xFF);

		readn = toWrite;

		if (readn) {
			size_t nwritten;
			nwritten = fwrite(buf, 1, readn, outf);
			if (nwritten != readn) {
				if (!quiet)
					fprintf(stderr, "Output file write error.\n");
				return 0;
			}
			ndone = readn;
		}
	} else {
		int foo = ftell(outf) - 8;

		fseek(outf, 4, SEEK_SET);
		fwrite(&foo,sizeof(foo),1,outf);

		foo = foo - pwf.datapos + 8;

		fseek(outf,pwf.datapos - sizeof(unsigned long), SEEK_SET);
		fwrite(&foo,sizeof(foo),1,outf);
		ndone = toWrite;
	}
	return ndone;
}

int process_file(char *fname)
{
	unsigned long	ndata;
	unsigned char	hdrbuf[16384];

	if (!quiet) {
		fprintf(stderr, "Processing file %s\n\n", fname);
	}

	// Open PCM WAV file
	if (!pcmwav_open(fname, "rb", &pwf)) {
		if (!quiet)
			fprintf(stderr, "%s\n", pcmwav_error);
		return 1;
	}

	if (nooverwrite) { // nooverwrite
		size_t	nread;
		unsigned int bits;
		fmt_sub* waveformat = ((fmt_sub*) (hdrbuf + pwf.formatpos));
		unsigned long *data_size = (unsigned long *)(hdrbuf + pwf.datasizepos);

		outf = fopen(outfname,"wb");
		if (outf == 0) {
			if (!quiet)
				fprintf(stderr, "Couldn't open output file '%s'.\n", outfname);
			return 1;
		}
		// Copy headers
		fseek(pwf.winfile, 0, SEEK_SET);
		nread = fread(hdrbuf, 1, pwf.datapos, pwf.winfile);
		if (nread != pwf.datapos) {
			if (!quiet)
				fprintf(stderr, "Could not copy headers.\n");
			return 1;
		}
		bits = waveformat->BitsPerSample;
		switch (bits) {
			case 1:
				waveformat->BitsPerSample = 8;
				*data_size = (*data_size) * 8;
				break;

			case 8:
				waveformat->BitsPerSample = 1;
				*data_size = (*data_size + 7) / 8;
				break;

			case 16:
				waveformat->BitsPerSample = 1;
				*data_size = (*data_size + 7) / 8;
				break;

			default:
				break;
		}
		waveformat->BlockAlign = (waveformat->BitsPerSample * waveformat->NumChannels) / 8;
		waveformat->ByteRate = (waveformat->BitsPerSample * waveformat->NumChannels * waveformat->SampleRate) / 8;
		fseek(pwf.winfile, pwf.datapos, SEEK_SET);
		nread = fwrite(hdrbuf, 1, pwf.datapos, outf);
		if (nread != pwf.datapos) {
			if (!quiet)
				fprintf(stderr, "Could not copy headers.\n");
			return 1;
		}
	}

	// Allocate buffer
	iobufsize = *((size_t *) (hdrbuf + 4));
	buf = malloc(iobufsize < pwf.filesize ? pwf.filesize : iobufsize);

	if (buf == NULL) {
		if (!quiet)
			fprintf(stderr, "Cannot allocate buffer in memory.\n");
		return 1;
	} else {
		if (!quiet)
			fprintf(stderr, "Allocated buffer size: %i.\n", iobufsize);
	}
	ndata = passthrough();

	free(buf);
	pcmwav_close(&pwf);

	if (nooverwrite)
		fclose(outf);

	return 0;
}

void usage(void) {
	fprintf(stderr, "\n%s\n", COPYRIGHT_NOTICE);
	fprintf(stderr,
		"    Usage:  wavconv [flags] input-file\n\n"

		"        -p           prompt before starting conversion\n"
		"        -o <file>    write output to <file> (instead of overwriting original)\n"
		"        -q           quiet (no screen output)\n"
		"        -d           don't abort batch if user skips the conversion of one file\n"
		"        -h           display this help\n\n"
		"        -t <value>   set comparison threshold to <value>\n\n"

		"    error levels: 0 = no error, 1 = I/O error, 2 = parameter error,\n"
		"                  3 = no conversion required, 4 = out of memory,\n"
		"                  5 = user abort\n\n"

		"	- wildcards are allowed in 'input-file' (e.g. wavconv *.wav)\n"
		"	- 'input-file' needs to be a PCM WAV file.\n");
}
